/*
 *--------------------------------------------------------------------------------------------
 * URLFetch (An ASP COM object written in Java that allows for the retrieval of files via HTTP)
 * Copyright (C) 2000 Todd Wilson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * The GNU General Public License may be found at the following URL:
 * http://www.gnu.org/copyleft/lgpl.html
 * 
 * Please also see the readme.txt file that accompanies this Visual J++ project.
 *
 * This is the source code for version 1.2 of URLFetch.
 *--------------------------------------------------------------------------------------------
 */

/**
 * Fetches URLs 'n stuff.
 * 
 * @author	Todd Wilson
 * @version	1.1.1
 */

import java.io.*;
import java.net.*;

/**
 * This class is designed to be packaged with a COM DLL output format.
 * The class has no standard entry points, other than the constructor.
 * Public methods will be exposed as methods on the default COM interface.
 * @com.register ( clsid=4B079286-7131-11D3-A432-0080AD38C188, typelib=4B079287-7131-11D3-A432-0080AD38C188 )
 */
public class URLFetch
{	
	/* Public variables */
	public String PostData = "";
	public String Proxy = "";
	public String ProxyPassword = "";
	
	/* Private variables */
	private String strURL = "";
	private String RequestHeaders = "";
	private boolean IsBinary = false;
	private byte[] bContent = null;
	private String[][] ResponseHeaders;	
	private StringBuffer Response;
	private String HTTPStatus = "";
	
	/** The number of milliseconds to delay when polling for the timeout. */
	private final int POLL_DELAY = 100;
	
	/** The connection timeout in milliseconds. */
	private long connectionTimeoutInterval = -1;
	
	/** The content received timeout in milliseconds. */
	private long contentReceivedTimeoutInterval = -1;
	
	/**
	 * The Fetch method is responsible for taking the URL, PostData, and any 
	 * RequestHeaders the user has set and issuing an HTTP request to the server
	 * indicated in the URL string.  It returns a string containing the body
	 * of the HTTP response.
	 */
	public String Fetch()
	{
		int i = 0;
		int beginPipe = 0;
		int endPipe = 0;
		String key = "";
		String value = "";
		String Cookies = "";
		String Cookie = "";
		String inputLine = "";
		Response = new StringBuffer();
		
		try 
		{	
			try
			{
				/*
				 * Beware!  These next two lines provide support for SSL fetches, but also taint URLFetch
				 * with a Microsoft "extension" to Java!
				 */
				Class clsFactory = Class.forName("com.ms.net.wininet.WininetStreamHandlerFactory" );
				URL.setURLStreamHandlerFactory((URLStreamHandlerFactory)clsFactory.newInstance());
			}
			catch (Exception e)
			{
				Response.append("Unable to load the Microsoft SSL stream handler. Check classpath." + e.toString());
			}
			catch (Error err)
			{
				/* Here we catch an error and eat it just in case the URLStreamHandlerFactory has already been set. */
			}
			
			URL url = new URL(strURL);
				
			try 
			{
				URLConnection connection = null;
				
				/*
				 * Here we set the proxy server (if the calling application has designated one),
				 * and, optionally, a proxy password.
				 * Thanks to Dave Walsh (paige386@yahoo.com) for adding proxy support.
				 */
				if (Proxy != "")
				{					
				  System.getProperties().put("proxySet", "true");
					System.getProperties().put("proxyHost", Proxy.substring(0,Proxy.indexOf(":")));
					System.getProperties().put("proxyPort", Proxy.substring(Proxy.indexOf(":")+1, Proxy.length()));	
					if (ProxyPassword != "")
					{
						connection.setRequestProperty("Proxy-Authorization", "Basic " + base64Encode(ProxyPassword));	
					}
				}
					
				connection = url.openConnection();
					
				/*
				 * Headers for the HTTP request are formatted in a pipe-delimited string
				 * in this way: |key1|value1|key2|value2|...
				 * Here we parse them out and either set them using the connection object,
				 * or, if they're cookies, put them all into the string Cookies, which
				 * is then used to set the Cookeis header all at once.  This allows us 
				 * to handle multiple cookies.
				 */
				RequestHeaders += "|";
					
				while (RequestHeaders.indexOf("|", beginPipe + 1) != -1)
				{						
					beginPipe = RequestHeaders.indexOf("|", beginPipe);
					endPipe = RequestHeaders.indexOf("|", beginPipe + 1);
					key = RequestHeaders.substring(beginPipe + 1, endPipe);
					beginPipe = endPipe;
						
					beginPipe = RequestHeaders.indexOf("|", beginPipe);
					endPipe = RequestHeaders.indexOf("|", beginPipe + 1);
					value = RequestHeaders.substring(beginPipe + 1, endPipe);
					beginPipe = endPipe;
						
					if (key.indexOf("Cookie") != -1)
					{
						Cookie = value;
						if (Cookie.indexOf(";") != -1)
							Cookie = Cookie.substring(0, Cookie.indexOf(";"));
						Cookies += Cookie + "; ";
					}
					else
					{
						connection.setRequestProperty(key, value);
					}
				}
				if (Cookies != "")
				{
					Cookies = Cookies.substring(0, Cookies.length() - 2);
					connection.setRequestProperty("Cookie", Cookies);
				}
				RequestHeaders = "";
					
				/*
				 * We check the length of PostData to see if the calling application has 
				 * added any data for a POST request.  If it has we tell the connection 
				 * that we'll be doing output, then we send it using a PrintWriter object.
				 */
				if (PostData.length() > 0) 
				{
					connection.setDoOutput(true);
					PrintWriter out = new PrintWriter(connection.getOutputStream());
					out.print(PostData);
					out.close();
				}
				
				
				if( connectionTimeoutInterval!=-1 )
				{
					ConnectThread connectThread = new ConnectThread( connection );
					connectThread.start();
					try
					{
						long startTime = System.currentTimeMillis();
						long pollTime = System.currentTimeMillis();
						
						while( pollTime < startTime + connectionTimeoutInterval )
						{
							Thread.sleep( POLL_DELAY );
							
							pollTime += POLL_DELAY;
							
							if( connectThread.isFinished() )
							{
								break;
							}
						}
						
						// If the we were never able to connect here we need to
						// fail.
						if( !connectThread.isFinished() )
						{
							return "Connection timed out.";
						}
					}
					catch( InterruptedException ie )
					{
						return "The connection thread was prematurely interrupted.";
					}
				
					i = connectThread.getNumHeaders();
				}
				else
				{
					i = getNumResponseHeaders( connection );
				}
									
				/*
				 * Once we know how many headers we've received in the HTTP response
				 * we declare a two-dimensional array (for keys and values) of the 
				 * appropriate size to store all of the headers.
				 */
				ResponseHeaders = new String[i][2];
					
				
				if( contentReceivedTimeoutInterval!=-1 )
				{
					ContentReceivedThread contentReceivedThread = new ContentReceivedThread( connection );
					contentReceivedThread.start();
					try
					{
						long startTime = System.currentTimeMillis();
						long pollTime = System.currentTimeMillis();
						
						while( pollTime < startTime + contentReceivedTimeoutInterval )
						{
							Thread.sleep( POLL_DELAY );
							
							pollTime += POLL_DELAY;
							
							if( contentReceivedThread.isFinished() )
							{
								break;
							}
						}
						
						// If the we were never able to connect here we need to
						// fail.
						if( !contentReceivedThread.isFinished() )
						{
							return "Timed out while retrieving content from the remote server.";
						}
						
						// If an error occurred just return it.
						if( contentReceivedThread.getErrorMessage()!=null )
						{
							return contentReceivedThread.getErrorMessage();
						}
					}
					catch( InterruptedException ie )
					{
						return "The content received thread was prematurely interrupted.";
					}					
				}
				else
				{
					// Read in all of the content.
					this.readContent( connection );
				}
				
				if( IsBinary )
				{
					return "binary";
				}
			}
				
			/*
			 * An IOException will be thrown if something goes wrong when the
			 * component tries to establish a connection to the remote server.
			 */
			catch (IOException e) 
			{
				Response.append(e.getMessage());
			}
		}
		/*
		 * A MalformedURLException is thrown if the URL isn't properly formatted.
		 * For example, if the user tries www.screen-scraper.com rather thatn http://www.screen-scraper.com/
		 */
		catch (MalformedURLException e) 
		{
			Response.append(e.getMessage());
		}		
		
		/*
		 * Return the Response string, which contains the body of the HTTP response.
		 * return Response.toString();
		 */
		return Response.toString();
	}
	
	
	/**
	 * Reads in all of the content from the remote server 
	 * using the given connection
	 * 
	 * @param	connection		The connection
	 */
	public void readContent( URLConnection connection ) throws IOException
	{
		int i = 0;
		String key = null;
		String value = null;
		String inputLine = null;
		
		/*
		 * Here we grab all of the headers in the HTTP response and put
		 * them into our ResponseHeaders array.
		 */
		while (connection.getHeaderField(i) != null)
		{
			key = connection.getHeaderFieldKey(i);
			key += "";
			ResponseHeaders[i][0] = key;
			value = connection.getHeaderField(i);
			value += "";
			ResponseHeaders[i][1] = value;
			i += 1;
		}
				
		/* Record the HTTP status. */
		HTTPStatus = connection.getHeaderField( 0 );

		if( IsBinary )
		{
			/* Read in the entire binary file in one big gulp. */
			DataInputStream in = new DataInputStream(connection.getInputStream());
			bContent = new byte[connection.getContentLength()];
			in.readFully(bContent);
			in.close();
		}
		else
		{
			/*
			 * Here we use the connection object to feed an InputStrem object to
			 * an InputStreamReader to a BufferedReader.  We use the BufferedReader
			 * object (in) to read in the HTTP response from the remote server.
			 */
			BufferedReader in = new BufferedReader(
												    new InputStreamReader(
															connection.getInputStream()));
				
			/*
			 * We read the body of the HTTP response into our Response string, 
			 * terminating each line with a newline character.
			 */
			while ((inputLine = in.readLine()) != null)
				Response.append(inputLine + "\n");
				
			/* Close down our BufferedReader object. */
			in.close();
		}
	}
	
	
	/**
	 * Gets the number of response headers from the given connection.
	 * 
	 * @param	connection		The connection.
	 * @return		The number of response headers.
	 */
	public int getNumResponseHeaders( URLConnection connection )
	{
		int numHeaders = 0;
		
		while (connection.getHeaderField(numHeaders) != null)
		{
			numHeaders += 1;
		}
		
		return numHeaders;
	}
	
	
	/**
	 * Used to allow for a timeout to be placed on the initial HTTP connection.
	 */
	private class ConnectThread extends TimeoutThread
	{
		private int numHeaders = 0;
					
		
		/**
		 * Initializes the class using the given connection.
		 */
		public ConnectThread( URLConnection connection )
		{
			super(connection);
		}
				
		
		/**
		 * Gets the number of headers contained in the HTTP
		 * response.
		 */
		public int getNumHeaders()
		{
			return this.numHeaders;
		}
		
		
		/**
		 * Retrieves the number of headers from the HTTP response.
		 */
		public void run()
		{
			this.numHeaders = getNumResponseHeaders( connection );
			
			this.isFinished = true;
		}
	}
	
	
	/**
	 * Used to allow for a timeout to be placed on receiving content 
	 * from the remote server.
	 */
	private class ContentReceivedThread extends TimeoutThread
	{		
		/** Will hold an error message, if one occurs. */
		String errorMessage = null;
		
		/**
		 * Initializes the class using the given connection.
		 */
		public ContentReceivedThread( URLConnection connection )
		{
			super(connection);
		}
		
		
		/**
		 * Returns an error message, if one occurred.
		 * 
		 * @return	The error message.
		 */
		public String getErrorMessage()
		{
			return this.errorMessage;
		}
		
		
		/**
		 * Retrieves the content from the remote server.
		 */
		public void run()
		{
			try
			{
				readContent( connection );
			}
			catch( IOException ioe )
			{
				this.errorMessage = "An error occurred while reading content from the server.  The message was: " + ioe.getMessage();
			}
			
			this.isFinished = true;
		}
	}
	
	
	/**
	 * Base class for the timeout thread classes.
	 */
	private abstract class TimeoutThread extends Thread
	{
		private URLConnection connection = null;
		private boolean isFinished = false;
					
		
		/**
		 * Initializes the class using the given connection.
		 */
		public TimeoutThread( URLConnection connection )
		{
			super();
			this.connection = connection;
		}
		
		
		/**
		 * Indicates whether or not the object has finished retrieving
		 * the number of HTTP response headers.
		 * 
		 * @return	A boolean.
		 */
		public boolean isFinished()
		{
			return this.isFinished;
		}
	}
	
	
	/*
	 * Use this method to set the URL to be fetched.  Note that this method is to 
	 * be used in lieu of the public URL variable, which was formerly used.
	 */
	public void SetURL(String URL)
	{
		strURL = URL;
	}
	
	
	/*
	 * GetHeaders is used to retrieve any headers that were received in the 
	 * HTTP request that was issued.
	 */
	public String[][] GetHeaders()
	{
		return ResponseHeaders;
	}
	
	
	/**
	 * The SetAuthorization method is used when access to the URL
	 * requires this type of authentication.
	 * Thanks to Real Gagnon (http://www.tactika.com/realhome/realhome.html) 
	 * for contributing the base64Encode code.
	 * @param	authorizationString	The <code>String</code> passed to the method should take 
	 *														the form username:passowrd.  The URLFetch object 
	 *														will properly encode the parameter in base64 format.
	 */
	public void SetAuthorization(String authorizationString)
	{
		String encodedString = "";
		
		encodedString = base64Encode(authorizationString);
		
		this.SetHeader("Authorization", "BASIC " + encodedString);
	}	
	
	/**
	 * Used to encode strings in base 64 format. 
	 * @param	toEncode	The <code>String</code> to be encoded.
	 */
	private String base64Encode(String toEncode)
	{
		String encodedString = "";
		byte bytes [] = toEncode.getBytes ();
		int i = 0;
		int pad = 0;				
		char base64Array [] = {
			'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/' 
		};
		
		while (i < bytes.length) {
			byte b1 = bytes [i++];
			byte b2;
			byte b3;
			if (i >= bytes.length) {
				b2 = 0;
				b3 = 0;
				pad = 2;
			}
			else {
				b2 = bytes [i++];
				if (i >= bytes.length) {           
					b3 = 0;
					pad = 1;
				}
				else
					b3 = bytes [i++];
			}
			byte c1 = (byte)(b1 >> 2);
			byte c2 = (byte)(((b1 & 0x3) << 4) | (b2 >> 4));
			byte c3 = (byte)(((b2 & 0xf) << 2) | (b3 >> 6));
			byte c4 = (byte)(b3 & 0x3f);
			encodedString += base64Array [c1];
			encodedString += base64Array [c2];
			switch (pad) {
			case 0:
				encodedString += base64Array [c3];
				encodedString += base64Array [c4];
				break;
			case 1:
				encodedString += base64Array [c3];
				encodedString += "=";
				break;
			case 2:
				encodedString += "==";
				break;
			}
		}
		return encodedString;	
	}
	
	/**
	 * The SetHeader method is used to set any headers that the calling
	 * program would like to issue along with the HTTP request.
	 * @param	Key	A <code>String</code> representing the name of the HTTP header 
	 *						to be set.
	 * @param	Value	A <code>String</code> representing the value of the HTTP header 
	 *							to be set.
	 */
	public void SetHeader(String Key, String Value)
	{
		RequestHeaders += "|" + Key + "|" + Value;	
	}
	
	/**
	 * Determines if the file to be fetched is binary. 
	 * @param	isBinary	A <code>boolean</code> indicating whether or not a binary
	 *									file is to be retrieved in this fetch. 
	 */
	public void SetIsBinary(boolean isBinary)
	{
		IsBinary = isBinary;	
	}
	
	/**
	 * Save the fetched contents as a file.  Binary files are handled
	 * differently from text-based files.
	 * @param	filePath	A <code>String</code> representing the local
	 *									system path (including the name of the file) that
	 *									the fetched contents should be saved as.
	 */
	public String SaveAsFile(String filePath)
	{
		if( IsBinary )
		{
			try
			{
				FileOutputStream fos = new FileOutputStream( filePath );
				fos.write( bContent );
				fos.close();
			}
			catch( IOException ioe )
			{
				return "An IOException occurred. The message was: " + ioe.getMessage();
			}
		}
		else
		{
			try
			{
				FileWriter fw = new FileWriter( filePath );
				fw.write( Response.toString() );
				fw.close();
			}
			catch( IOException ioe )
			{
				return "An IOException occurred. The message was: " + ioe.getMessage() + ioe.getMessage() + ".";;
			}
			catch( NullPointerException npe )
			{
				return "A NullPointerException occurred.  The message was: " + npe.getMessage() + ".";
			}
		}
		
		return "success";
	}
	
	/**
	 * If a binary file was fetched, this function will return it as a byte array. 
	 */
	public com.ms.com.Variant GetBinary()
	{
		/*
		 * ARG! Yes, I know another Microsoft "extension".  Well, this
		 * is the only way to translate a Java byte array into something
		 * COM can deal with.  So just deal with it.
		 */
		com.ms.com.SafeArray saBytes = new com.ms.com.SafeArray(com.ms.com.Variant.VariantByte, bContent.length);
		saBytes.fromByteArray(bContent);
		
		return new com.ms.com.Variant(saBytes, true);
	}
	
	/**
	 * Gets a header by name. 
	 * @param	name	A <code>String</code> representing the name of the
	 *							HTTP header to be returned.  Note that this function
	 *							should be used _after_ a fetch has been done, as it
	 *							will return the HTTP headers associated with the 
	 *							fetched file.
	 */
	public String GetHeader( String name )
	{
		for( int i = 0; i < ResponseHeaders.length; i++ )
		{
			if( name.equals( (String) ResponseHeaders[i][0] ) )
			{
				return (String) ResponseHeaders[i][1];
			}
		}
		
		return "";
	}
	
	/**
	 * Returns the HTTP status of a response.
	 */
	public String GetHTTPStatus()
	{
		return HTTPStatus;
	}
	
	/**
	 * Sets the amount of time the component should wait before
	 * giving up on trying to connect to a remote server.
	 * 
	 * @param	connectionTimeoutInterval	The time in milliseconds.
	 */
	public void SetConnectionTimeout( long connectionTimeoutInterval )
	{
		this.connectionTimeoutInterval = connectionTimeoutInterval;
	}
	
	/**
	 * Sets the amount of time the component should wait before
	 * giving up on receiving all content from a remote server.
	 * 
	 * @param	contentReceivedTimeoutInterval	The time in milliseconds.
	 */
	public void SetContentReceivedTimeout( long contentReceivedTimeoutInterval )
	{
		this.contentReceivedTimeoutInterval = contentReceivedTimeoutInterval;
	}
}
